@@ -1,5 +1,6 @@
 from datetime import timedelta
-from typing import Any, Callable, Dict, List, Literal, Tuple
+from functools import reduce
+from typing import Any, Callable, Dict, List, Tuple
 
 import pandas as pd
 from pandas import DataFrame
@@ -15,7 +16,6 @@ from seshat.data_class.pandas import DFrame
 from seshat.general import configs
 from seshat.general.exceptions import InvalidArgumentsError
 from seshat.transformer import Transformer
-from seshat.transformer.merger import Merger
 from seshat.transformer.schema import Col, Schema
 from seshat.transformer.trimmer.base import FilterTrimmer
 from seshat.utils import pandas_func, pyspark_func
@@ -257,7 +257,7 @@ class OperationOnColsDeriver(SFrameDeriver):
                 self.agg_func, axis=1
             )
         else:
-            default[self.result_col] = default.apply(self.agg_func, axis=1)
+            default[self.result_col] = default[self.cols].apply(self.agg_func, axis=1)
 
         return {"default": default}
 
@@ -265,14 +265,19 @@ class OperationOnColsDeriver(SFrameDeriver):
         if self.is_numeric:
             for col in self.cols:
                 default = NumericColumnValidator().validate(default, col)
-        try:
-            func = pyspark_func.func_maps[self.agg_func]
-        except KeyError:
-            raise KeyError(
-                "func %s not available for pyspark dataframe" % self.agg_func
-            )
+        if isinstance(self.agg_func, str):
+            try:
+                func = pyspark_func.func_maps[self.agg_func]
+            except KeyError:
+                raise KeyError(
+                    "func %s not available for pyspark dataframe" % self.agg_func
+                )
+        else:
+            func = self.agg_func
         func_udf = F.udf(func)
-        default = default.withColumn(self.result_col, func_udf(*self.cols))
+        default = default.withColumn(
+            self.result_col, func_udf(*[F.col(col) for col in self.cols])
+        )
         if self.is_numeric:
             default = NumericColumnValidator().validate(default, self.result_col)
         return {"default": default}
@@ -1090,140 +1095,6 @@ class GroupByDeriverMeanMax(SFrameDeriver):
         return {"default": default}
 
 
-class TimeWindowTransformer(SFrameDeriver):
-    """
-    A class that derives time series-based features from a default SFrame using a transformer and merger strategy.
-    It supports applying transformations over multiple durations and combines the results using a merger.
-
-    Duration labeling can be handled in two ways:
-    1. By adding a new column with the name specified in `label_col`, where each row is labeled with its corresponding duration.
-    2. By appending the duration label as a suffix to the names of the new columns created by the transformer, excluding those required for merging.
-
-    This behavior is controlled by the `specify_duration_by` parameter.
-
-    Parameters
-    ----------
-    transformer : Transformer or Callable
-        An instance of a transformer or a callable function that performs the transformation on the input data.
-    merger : Merger
-        An instance of a merger used to combine features generated across different durations.
-    time_col : str, optional
-        The name of the timestamp column in the SFrame. Defaults to `configs.BLOCK_TIMESTAMP_COL`.
-    durations : List[Tuple[str, int]]
-        A list of (label, duration) pairs where each label identifies a time window (in minutes) for feature extraction.
-    label_col : str, optional
-        The name of the column that will store the duration label for each record. Defaults to `configs.DURATION_LABEL_COL`.
-    specify_duration_by : {'column', 'suffix'}, default 'column'
-        Specifies how duration information is represented in the output:
-        * 'column' – Adds a new column (named by `label_col`) containing the duration label.
-        * 'suffix' – Appends the duration label as a suffix to the names of columns created by the transformer,
-          excluding those required by the merger.
-    """
-
-    DEFAULT_GROUP_KEYS = {
-        "default": configs.DEFAULT_SF_KEY,
-        "timeseries_feature": configs.TIMESERIES_SF_KEY,
-    }
-
-    def __init__(
-        self,
-        transformer: Transformer | Callable,
-        merger: Merger,
-        durations: List[Tuple[str, int]],
-        time_col=configs.BLOCK_TIMESTAMP_COL,
-        label_col=configs.DURATION_LABEL_COL,
-        specify_duration_by: Literal["column", "suffix"] = "column",
-        group_keys=None,
-    ):
-        super().__init__(group_keys)
-        self.time_col = time_col
-        self.transformer = transformer
-        self.label_col = label_col
-        self.merger = merger
-        self.specify_duration_by = specify_duration_by
-        self.durations = durations
-
-        # Ensure that the merge works in place and group keys are set to their default values.
-        self.merger.inplace = True
-        self.merger.group_keys = {"default": "default", "other": "other"}
-        if self.specify_duration_by == "column":
-            self.merger.axis = 0
-
-    def validate(self, sf: SFrame):
-        super().validate(sf)
-        self._validate_columns(sf, self.group_keys["default"], self.time_col)
-
-    def _is_empty(self, sf: SFrame) -> bool:
-        raw = sf.to_raw()
-        if raw is None:
-            return True
-        elif hasattr(raw, "rdd"):
-            return raw.rdd.isEmpty()
-        else:
-            return len(raw) == 0
-
-    def calculate_metrics(self, sf: SFrame, label, start) -> SFrame:
-        trimmer = FilterTrimmer(op=">=", col=self.time_col, value=start)
-        filtered = trimmer(sf)
-
-        if self._is_empty(filtered):
-            return filtered
-
-        result = self.transformer(sf_input=filtered, label=label, start=start)
-        if self.specify_duration_by == "column":
-            deriver = StaticValueColumnAdder(col_name=self.label_col, value=label)
-            result = deriver(result)
-
-        else:
-            # Find columns excluding those used by,
-            # the merger for merging.
-            cols = set(result[self.transformer].get_columns()) - set(self.merger.on)
-            schema = Schema(
-                cols=[Col(col, to=f"{col}_{label}") for col in cols],
-                exclusive=False,
-            )
-            result = schema(result)
-        return result
-
-    def derive_df(self, default: DataFrame, *args, **kwargs):
-        default = TimeStampColumnValidator().validate(default, self.time_col)
-        last_timestamp = default[self.time_col].max()
-        sf_input = DFrame.from_raw(default)
-        result = DFrame()
-        return self.derive(sf_input, result, last_timestamp)
-
-    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
-        default = TimeStampColumnValidator().validate(default, self.time_col)
-        last_timestamp = default.select(F.max(self.time_col)).collect()[0][0]
-        sf_input = SPFrame.from_raw(default)
-        result = SPFrame()
-        return self.derive(sf_input, result, last_timestamp)
-
-    def derive(self, sf_input: SFrame, result: SFrame, last_timestamp):
-        intervals = {
-            label: last_timestamp - timedelta(minutes=minutes)
-            for label, minutes in self.durations
-        }
-
-        for idx, (label, start) in enumerate(intervals.items()):
-            metrics = self.calculate_metrics(sf_input, label, start)
-            if self._is_empty(metrics):
-                continue
-            metrics = metrics.make_group()
-            metrics["other"] = metrics["default"]
-            metrics["default"] = result
-            result = (
-                metrics["other"]
-                if self._is_empty(result)
-                else self.merger(metrics)["default"]
-            )
-
-        return {
-            "default": sf_input[self.default_sf_key].to_raw(),
-            "timeseries_feature": result.to_raw(),
-        }
-
-
 class StaticValueColumnAdder(SFrameDeriver):
     """
     A deriver that adds a new column with a static value to a DataFrame.
@@ -1564,7 +1435,6 @@ class ComprehensiveFeaturesDeriver(SFrameDeriver):
         **kwargs,
     ):
         def calculate_features(df):
-
             # Group by TOKEN_ADDRESS to calculate features
             features = (
                 df.groupby("TOKEN_ADDRESS")
@@ -1882,13 +1752,11 @@ class TokenFeatureTransformationDeriver(SFrameDeriver):
         df = token_price_df.copy()
 
         for interval_label in self.time_intervals_price["ago_labels"]:
-
             df[f"Price_Change_Percentage_{interval_label}"] = (
                 df["CURRENT_PRICE"] - df[f"PRICE_{interval_label}"]
             ) / df[f"PRICE_{interval_label}"]
 
         for interval_label in self.time_intervals_trade["last_labels"]:
-
             df[f"Buy_to_Sell_Volume_Ratio_{interval_label}"] = df[
                 f"VOLUME_INFLOW_{interval_label}"
             ] / (df[f"VOLUME_OUTFLOW_{interval_label}"] + epsilon)
@@ -1937,3 +1805,393 @@ class TokenFeatureTransformationDeriver(SFrameDeriver):
     def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
         # ToDo: add spf
         return {"default": default}
+
+
+class BranchClassifier(SFrameDeriver):
+    """
+    BranchClassifier takes an SFrame and generates multiple SFrames based on grouping by specified columns.
+
+    Parameters
+    ----------
+    group_cols : list of str
+        List of column names to group by. Each unique combination will create a separate SFrame.
+    group_keys : dict, optional
+        Dictionary mapping group names to their respective keys. If None, will auto-generate
+        keys based on group column values.
+    prefix : str, optional
+        Prefix for auto-generated group keys. Defaults to "".
+    keep_group_cols : bool, optional
+        Whether to keep the group columns in each resulting SFrame. Defaults to False.
+        When False, the group columns are removed from each grouped SFrame since they
+        contain the same value for all rows in that group.
+    keep_unhandled_sframes : bool, optional
+        If True (default), any child SFrames of the input SFrame that are not handled by this classifier
+        will be preserved in the output. If False, all unhandled child SFrames will be removed from the output,
+        and only the newly generated grouped SFrames will be present.
+    """
+
+    ONLY_GROUP = False
+    DEFAULT_GROUP_KEYS = {"default": configs.DEFAULT_SF_KEY}
+
+    def __init__(
+        self,
+        group_cols: List[str],
+        group_keys=None,
+        prefix: str = "",
+        keep_group_cols: bool = False,
+        keep_unhandled_sframes: bool = True,
+        *args,
+        **kwargs,
+    ):
+        super().__init__(group_keys, *args, **kwargs)
+        self.group_cols = group_cols
+        self.prefix = prefix
+        self.keep_group_cols = keep_group_cols
+        self.keep_unhandled_sframes = keep_unhandled_sframes
+
+    def calculate_complexity(self):
+        return len(self.group_cols)
+
+    def execute(self, sf_input: SFrame, *args, **kwargs):
+        result = self.call_handler(sf_input, *args, **kwargs)
+        sf_output = sf_input.make_group(self.default_sf_key)
+        if not self.keep_unhandled_sframes:
+            sf_output.children.clear()
+
+        self.set_raw(sf_output, result)
+        return sf_output
+
+    def validate(self, sf: SFrame):
+        super().validate(sf)
+        self._validate_columns(sf, self.default_sf_key, *self.group_cols)
+
+    def derive_df(self, default: DataFrame, *args, **kwargs) -> Dict[str, DataFrame]:
+        result = {}
+
+        # Group by columns and create separate DataFrames
+        for group_values, group_df in default.groupby(self.group_cols):
+            if not isinstance(group_values, tuple):
+                group_values = (group_values,)
+
+            group_key_parts = []
+            for value in group_values:
+                if pd.isna(value):
+                    group_key_parts.append("null")
+                else:
+                    group_key_parts.append(str(value).replace(" ", "_"))
+
+            group_key = self.prefix + "_".join(group_key_parts)
+
+            clean_df = group_df.reset_index(drop=True)
+            if not self.keep_group_cols:
+                clean_df = clean_df.drop(columns=self.group_cols)
+
+            result[group_key] = clean_df
+
+        return result
+
+    def derive_spf(
+        self, default: PySparkDataFrame, *args, **kwargs
+    ) -> Dict[str, PySparkDataFrame]:
+        result = {}
+        non_null_df = default.dropna(subset=self.group_cols)
+        unique_groups = non_null_df.select(*self.group_cols).distinct().collect()
+
+        for row in unique_groups:
+            conditions = [F.col(col) == row[col] for col in self.group_cols]
+            group_key = self.prefix + "_".join(
+                str(row[col]).replace(" ", "_") for col in self.group_cols
+            )
+
+            group_df = default.filter(reduce(lambda a, b: a & b, conditions))
+            if not self.keep_group_cols:
+                group_df = group_df.drop(*self.group_cols)
+
+            result[group_key] = group_df
+        return result
+
+
+class Tagger(SFrameDeriver):
+    """
+    This class is used to apply multiple filters to data and attach labels to rows
+    that pass each filter. A row can have multiple labels if it passes multiple filters.
+
+    Parameters
+    ----------
+    filters_and_labels : list of tuple
+        List of tuples containing (FilterTrimmer, label) pairs. Each FilterTrimmer
+        will be applied and rows that pass the filter will get the corresponding label.
+    result_col : str, default "labels"
+        Name of the column to store the list of labels for each row.
+    """
+
+    def __init__(
+        self,
+        filters_and_labels: List[Tuple[FilterTrimmer, str]],
+        result_col: str = "labels",
+        group_keys=None,
+    ):
+        super().__init__(group_keys)
+        self.filters_and_labels = filters_and_labels
+        self.result_col = result_col
+
+    def calculate_complexity(self):
+        return len(self.filters_and_labels) * 5
+
+    def validate(self, sf: SFrame):
+        super().validate(sf)
+        # Validate that all filters can work with the input data
+        for filter_trimmer, _ in self.filters_and_labels:
+            filter_trimmer.validate(sf)
+
+    def derive_df(self, default: DataFrame, *args, **kwargs) -> Dict[str, DataFrame]:
+        results = []
+        for filter_trimmer, label in self.filters_and_labels:
+            res = filter_trimmer.trim_df(default, *args, **kwargs)["default"].copy()
+            res.loc[:, self.result_col] = label
+            results.append(res)
+
+        result = pd.concat(results)
+
+        without_label = result[
+            [col for col in result.columns if col != self.result_col]
+        ]
+        unlabeled = pd.concat([default, without_label]).drop_duplicates(keep=False)
+
+        result = pd.concat([result, unlabeled])
+
+        return {"default": result}
+
+    def derive_spf(
+        self, default: PySparkDataFrame, *args, **kwargs
+    ) -> Dict[str, PySparkDataFrame]:
+        results = []
+
+        for filter_trimmer, label in self.filters_and_labels:
+            filtered_result = filter_trimmer.trim_spf(default, *args, **kwargs)
+            res = filtered_result["default"]
+
+            res = res.withColumn(self.result_col, F.lit(label))
+            results.append(res)
+
+        if results:
+            result = results[0]
+            for df in results[1:]:
+                result = result.union(df)
+        else:
+            result = default.withColumn(self.result_col, F.lit(None).cast("string"))
+
+        # Find unlabeled rows
+        other_columns = [col for col in default.columns if col != self.result_col]
+        without_label = result.select(*other_columns)
+        unlabeled = default.join(without_label, on=other_columns, how="left_anti")
+        unlabeled = unlabeled.withColumn(self.result_col, F.lit(None).cast("string"))
+        final_result = result.union(unlabeled)
+
+        return {"default": final_result}
+
+
+class ProductSumDeriver(SFrameDeriver):
+    """
+    Calculates weighted sum of multiple columns and stores result in a new column.
+
+    Parameters
+    ----------
+    feature_weights : List[Tuple[str, float | int]]
+        List of (column_name, weight) pairs for weighted sum calculation
+    result_col : str
+        Name of the column to store the weighted sum result
+    """
+
+    def __init__(
+        self,
+        feature_weights: List[Tuple[str, float | int]],
+        result_col: str,
+        group_keys=None,
+        *args,
+        **kwargs,
+    ):
+        super().__init__(group_keys, *args, **kwargs)
+        self.feature_weights = feature_weights
+        self.result_col = result_col
+
+    def validate(self, sf: SFrame):
+        super().validate(sf)
+        self._validate_columns(
+            sf, self.default_sf_key, *[c for c, _ in self.feature_weights]
+        )
+
+    def derive_df(self, default: DataFrame, *args, **kwargs):
+        default[self.result_col] = sum(
+            [weight * default[col] for col, weight in self.feature_weights]
+        )
+        return {"default": default}
+
+    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
+        weighted_cols = [
+            F.lit(weight) * F.col(col_name) for col_name, weight in self.feature_weights
+        ]
+        default = default.withColumn(
+            self.result_col, reduce(lambda x, y: x + y, weighted_cols)
+        )
+        return {"default": default}
+
+
+class RankDeriver(SFrameDeriver):
+    """
+    Derives a new column with dense rank values based on specified column.
+
+    Parameters
+    ----------
+    col : str
+        Column to calculate ranks from
+    result_col : str
+        Column to store the calculated ranks
+    ascending : bool, default False
+        Whether to rank in ascending order
+    """
+
+    def __init__(
+        self,
+        col: str,
+        result_col: str,
+        ascending: bool = False,
+        group_keys=None,
+        *args,
+        **kwargs,
+    ):
+        super().__init__(group_keys, *args, **kwargs)
+        self.col = col
+        self.result_col = result_col
+        self.ascending = ascending
+
+    def validate(self, sf: SFrame):
+        super().validate(sf)
+        self._validate_columns(sf, self.default_sf_key, self.col)
+
+    def derive_df(self, default: DataFrame, *args, **kwargs):
+        default[self.result_col] = (
+            default[self.col].rank(ascending=self.ascending, method="dense").astype(int)
+        )
+        return {"default": default}
+
+    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
+        orderby_col = F.col(self.col)
+        orderby_col = orderby_col.asc() if self.ascending else orderby_col.desc()
+        window = Window.orderBy(orderby_col)
+        default = default.withColumn(self.result_col, F.dense_rank().over(window))
+        return {"default": default}
+
+
+class ShiftDeriver(SFrameDeriver):
+    """
+    Derives new columns by applying shifts to a specified column, grouped by given keys.
+
+    Parameters
+    ----------
+    sort_on : List[str]
+        Columns to sort on before applying shifts
+    shifts : List[Tuple[str, int]]
+        List of (label, shift_value) pairs defining each shift operation
+    col : str
+        Column to shift values from
+    group_by : str | List[str]
+        Column(s) to group by when applying shifts
+    ascending : bool, default True
+        Sort order direction
+    """
+
+    def __init__(
+        self,
+        sort_on: List[str],
+        shifts: List[Tuple[str, int]],
+        col: str,
+        group_by: str | List[str],
+        ascending: bool = True,
+        group_keys=None,
+        *args,
+        **kwargs,
+    ):
+        super().__init__(group_keys, *args, **kwargs)
+        self.sort_on = sort_on
+        self.ascending = ascending
+        self.shifts = shifts
+        self.col = col
+        if isinstance(group_by, str):
+            group_by = [group_by]
+        self.group_by = group_by
+
+    def validate(self, sf: SFrame):
+        super().validate(sf)
+        self._validate_columns(sf, self.default_sf_key, self.col)
+        self._validate_columns(sf, self.default_sf_key, *self.group_by)
+
+    def derive_df(self, default: DataFrame, *args, **kwargs):
+        default = default.sort_values(
+            by=self.sort_on, ascending=self.ascending
+        ).reset_index(drop=True)
+
+        for label, shift in self.shifts:
+            default[f"{self.col}_{label}"] = default.groupby(self.group_by)[
+                self.col
+            ].shift(shift)
+
+        default = default.groupby(self.group_by)
+        default = default.last() if self.ascending else default.first()
+        default = default.reset_index()
+
+        return {"default": default}
+
+    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
+        default = default.orderBy(
+            [
+                F.col(c).asc() if self.ascending else F.col(c).desc()
+                for c in self.sort_on
+            ]
+        )
+        window_spec = Window.partitionBy(self.group_by).orderBy(self.sort_on)
+        for label, shift in self.shifts:
+            if shift > 0:
+                default = default.withColumn(
+                    f"{self.col}_{label}",
+                    F.lag(F.col(self.col), shift).over(window_spec),
+                )
+            elif shift < 0:
+                default = default.withColumn(
+                    f"{self.col}_{label}",
+                    F.lead(F.col(self.col), abs(shift)).over(window_spec),
+                )
+            else:
+                default = default.withColumn(f"{self.col}_{label}", F.col(self.col))
+
+        if self.ascending:
+            default = default.groupBy(self.group_by).agg(
+                *[F.last(c).alias(c) for c in default.columns if c not in self.group_by]
+            )
+        else:
+            default = default.groupBy(self.group_by).agg(
+                *[
+                    F.first(c).alias(c)
+                    for c in default.columns
+                    if c not in self.group_by
+                ]
+            )
+
+        return {"default": default}
+
+
+class DuplicatedSFrameDeriver(SFrameDeriver):
+    """
+    Duplicate an SFrame and add it to the group sframe with a new key.
+    """
+
+    DEFAULT_GROUP_KEYS = {
+        "default": configs.DEFAULT_SF_KEY,
+        "duplicated": configs.DUPLICATED_SF_KEY,
+    }
+
+    def derive_df(self, default: DataFrame, *args, **kwargs):
+        return {"default": default, "duplicated": default.copy()}
+
+    def derive_spf(self, default: PySparkDataFrame, *args, **kwargs):
+        return {"default": default, "duplicated": default.select("*")}
